// -*- C++ -*-
// --------------------------------------------------------------------
// AppUi
// --------------------------------------------------------------------
/*

    This file is part of the extensible drawing editor Ipe.
    Copyright (C) 1993-2007  Otfried Cheong

    Ipe is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    As a special exception, you have permission to link Ipe with the
    CGAL library and distribute executables, as long as you follow the
    requirements of the Gnu General Public License in regard to all of
    the software in the executable aside from CGAL.

    Ipe is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
    License for more details.

    You should have received a copy of the GNU General Public License
    along with Ipe; if not, you can find it at
    "http://www.gnu.org/copyleft/gpl.html", or write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#ifndef APPUI_H
#define APPUI_H

#include "ipebase.h"
#include "ipecolor.h"
#include "ipestyle.h"
#include "ipesnap.h"
#include "ipelet.h"
#include "ipeundo.h"

#include "ipemodel.h"
#include "ipeprefs.h"

#include "ipecanvas.h"
#include "ipeoverlay.h"

#include "master.h"

#include <QtGui>

#ifdef MessageBox
#undef MessageBox
#endif

class IpePage;
class IpeStyleSheet;
class ZoomSpinBox;
class IpeSpinBox;
class ArrowButton;
class DecimalSpinBox;
class LayerBox;

// --------------------------------------------------------------------

class AppUi : public QMainWindow, public IpeCanvasServices,
	      public IpeOverlayServices, public IpeletHelper {
  Q_OBJECT

public:
  enum TCommands { ECmdGroup, ECmdUngroup, ECmdDuplicate, ECmdFront, ECmdBack,
		   ECmdForward, ECmdBackward, ECmdBehind, ECmdBefore,
		   ECmdCut, ECmdCopy, ECmdPaste, ECmdDelete, ECmdSelectAll,
		   ECmdNewLayer, ECmdLineWidth, ECmdDashStyle,
		   ECmdTextSize, ECmdMarkShape, ECmdMarkSize,
		   ECmdTextStyle, ECmdVerticalAlign, ECmdHorizontalAlign,
		   ECmdTransformableText, ECmdPinned,
		   ECmdLineJoin, ECmdLineCap, ECmdWindRule,
		   ECmdUndo, ECmdRedo, ECmdEdit,
		   ECmdSetArrows, ECmdArrowSize, ECmdStroke, ECmdFill,
		   ECmdComposePaths, ECmdJoinPaths, ECmdDecomposePath,
		   ECmdHere, ECmdSetOrigin, ECmdResetOrigin,
		   ECmdSetOriginSnap, ECmdSetDirection, ECmdSetDirectionSnap,
		   ECmdResetDirection,
		   ECmdSetLine, ECmdSetLineSnap,
		   ECmdNewLayerNewView, ECmdNewView, ECmdDeleteView,
		   ECmdFirstView, ECmdLastView,
  };

public:
  AppUi(QWidget* parent=0, Qt::WFlags f=0);
  ~AppUi();

  bool Load(QString fileName);
  bool Save(QString fname, bool warnOverwrite);

  static void ReadKeyAssignment();

public slots:
  void Cmd(int cmd);
  void UndoCmd(int cmd);
  void SnapCmd(int cmd);

  void InsertTextBox();

  void NewWindow();

  void About();
  void AboutIpelets();

  void NewDoc();
  bool Save();
  bool SaveAs();
  void SaveAsBitmap();
  void Load();

  void EditObject();
  void changeTextWidth();
  void EditDocProps();
  void EditPrefs();
  void Manual();
  bool RunLatex(bool force = true);

  void NextView();
  void PreviousView();
  void EditBookmarks();
  void EditEffects();

  void CreatePage();
  void DeletePage();
  void CopyPage(bool cut = false);
  void CutPage();
  void PastePage();

  void NormalSize();
  void DefaultZoom();
  void FitPage();
  void FitObjects();
  void FitSelection();

  void StyleSheets();
  void UpdateStyleSheets();

  void FirstPage();
  void LastPage();

  void SetCreator(QAction *action);
  void SetMouseDisplayIn(int);

  void RunIpelet(int);

protected:
  void closeEvent(QCloseEvent *ev);

private slots:
  void aboutToShowEditMenu();
  void aboutToShowPropertiesMenu();
  void aboutToShowMouseUnitsMenu();

  void PageChanged(int);
  void PageChanged();
  void ViewChanged(int);
  void ViewChanged();
  void ResolutionChanged(int);
  void AbsoluteAttributes(bool);
  void AbsoluteSnapping(bool);
  void ToggleAbsoluteAttributes();

  void SetStrokeColor();
  void SetFillColor();
  void SetStrokeColorName(int item);
  void SetFillColorName(int item);

  void setMarkShape(int id);
  void setTransformable(int id);
  void setHorizontalAlignment(int id);
  void setVerticalAlignment(int id);
  void setTextStyle(int id);
  void setPinned(int id);
  void setLineJoin(int id);
  void setLineCap(int id);
  void setWindRule(int id);

  void moveToLayer(int id);
  void selectInLayer(int id);

  void LineWidthChanged(int id);
  void AbsLineWidthChanged(int val);
  void DashStyleChanged(int id);
  void TextSizeChanged(int id);
  void AbsTextSizeChanged(int val);
  void MarkSizeChanged(int id);
  void AbsMarkSizeChanged(int val);
  void ArrowSizeChanged(int id);
  void AbsArrowSizeChanged(int val);
  void ArrowChanged(bool on);
  void GridSizeChanged(int id);
  void AbsGridSizeChanged(int val);
  void AngleSizeChanged(int id);
  void AbsAngleSizeChanged(int val);

  void SnapChanged(bool);
  void GridVisible(bool);
  void CoordinatesVisible(bool);
  void LayerRightPress(QListWidgetItem *item, const QPoint &pos);
  void LayerChanged(QListWidgetItem *item);
  void BookmarkSelected(QListWidgetItem *item);

  void PreferencesChanged();

private: // from IpeCanvasServices
  virtual IpeBuffer StandardFont(IpeString fontName);
  virtual void CvSvcRequestOverlay(QMouseEvent *ev);
  virtual void CvSvcSetDrawingMode(bool mode);
  virtual void CvSvcWheelZoom(int delta);
  virtual void CvSvcMousePosition(const IpeVector &pos);
  virtual void CvSvcClearMessage();

private: // from IpeOverlayServices
  virtual void OvSvcAddObject(IpeObject *obj);
  virtual const IpeAllAttributes &OvSvcAttributes();
  virtual const IpeStyleSheet *OvSvcStyleSheet();
  virtual void OvSvcAddUndoItem(IpePage *page, QString s);
  virtual void OvSvcAddUndoItem(IpePage::iterator it, IpePage *page,
				const IpePgObject &original, QString s);
  virtual void OvSvcMessage(QString s);

private: // from IpeletHelper
  virtual void Message(const char *msg);
  virtual int MessageBox(const char *text, const char *button1,
			 const char *button2, const char *button3);
  virtual bool GetString(const char *prompt, IpeString &str);
  virtual const IpeStyleSheet *StyleSheet();
  virtual const IpeDocument *Document();
  virtual IpeDocument *EditDocument();
  virtual int CurrentPage() const;
  virtual int CurrentView() const;
  virtual int CurrentLayer() const;
  virtual const IpeSnapData &SnapData() const;
  virtual const IpeAllAttributes &Attributes() const;

private:
  QKeySequence Key(IpeString function) const;
  inline IpePage *Page() { return iDoc->page(iPno); }
  void FitBox(const IpeRect &box);
  void UpdateLayers();
  void SetGridSizeFromPage();
  void SwitchAttributeUi();
  void SwitchSnapUi();
  void SetCaption();
  void ShowStyleInUi();
  void AddUndo(IpeUndoItem *item);
  void UpdateUndo();
  IpeAttribute PageColor();

  void AutoExport(IpeDocument::TFormat fm, const QString &fn,
		  bool onlyIfExists);

  void ErrMsg(QString str);

  IpeFixed AbsValue(IpeAttribute attr);
  void ConvertAbsolute(DecimalSpinBox *spin, IpeAttribute &attr);

  void ResetCombo(IpeKind kind, QComboBox *combo,
		  IpeAttribute &attr, QPixmap *pixmap = 0);
  void InsertColor(QComboBox *combo, IpeAttribute sym);
  void SetColorFromQColor(IpeAttribute &colorToSet, QColor &qColor,
			  QPushButton *button);

  virtual bool eventFilter(QObject *obj, QEvent *ev);
  void addActions(QMenu *menu, const char *text[], int num, const char *slot);

private:
  // UI components
  IpeCanvas *iCanvas;

  QMenu *iFileMenu;
  QMenu *iEditMenu;
  QMenu *iPropertiesMenu;
  QMenu *iSnapMenu;
  QMenu *iModeMenu;
  QMenu *iZoomMenu;
  QMenu *iLayerMenu;
  QMenu *iViewMenu;
  QMenu *iPageMenu;
  QMenu *iIpeletMenu;
  QMenu *iHelpMenu;

  QMenu *iTransformableMenu;
  QMenu *iTextStyleMenu;
  QMenu *iMarkShapeMenu;
  QMenu *iLabelVerticalAlignmentMenu;
  QMenu *iLabelHorizontalAlignmentMenu;
  QMenu *iPinnedMenu;
  QMenu *iLineJoinMenu;
  QMenu *iLineCapMenu;
  QMenu *iWindRuleMenu;

  QMenu *iMouseUnitsMenu;

  QMenu *iSelectInLayerMenu;
  QMenu *iMoveToLayerMenu;

  QToolBar *iFileTools;
  QToolBar *iResolutionTools;
  QToolBar *iPageTools;
  QToolBar *iSnapTools;
  QToolBar *iObjectTools;

  QDockWidget *iPropertiesTools;
  QDockWidget *iLayerTools;
  QDockWidget *iBookmarkTools;

  QStackedWidget *iStrokeColorStack;
  QStackedWidget *iFillColorStack;
  QStackedWidget *iLineWidthStack;
  QStackedWidget *iArrowSizeStack;
  QStackedWidget *iTextSizeStack;
  QStackedWidget *iMarkSizeStack;
  QStackedWidget *iGridSizeStack;
  QStackedWidget *iAngleSizeStack;

  QAction *iCutAction;
  QAction *iCopyAction;
  QAction *iDeleteAction;
  QAction *iGroupAction;
  QAction *iFrontAction;
  QAction *iBackAction;
  QAction *iDuplicateAction;

  QAction *iUngroupAction;
  QAction *iComposeAction;
  QAction *iJoinAction;

  QAction *iAbsoluteAttributesAction;
  QAction *iUndoAction;
  QAction *iRedoAction;
  QAction *iGridVisibleAction;
  QAction *iCoordinatesVisibleAction;
  QAction *iAbsoluteSnapAction;

  QSpinBox *iResolution;

  IpeSpinBox *iPageNumber;
  IpeSpinBox *iViewNumber;

  QComboBox *iStrokeColor;
  QPushButton *iAbsStrokeColor;
  QComboBox *iFillColor;
  QPushButton *iAbsFillColor;

  QComboBox *iLineWidth;
  DecimalSpinBox *iAbsLineWidth;
  QComboBox *iDashStyle;
  ArrowButton *iBackwardArrow;
  ArrowButton *iForwardArrow;
  QComboBox *iArrowSize;
  DecimalSpinBox *iAbsArrowSize;

  QActionGroup *iModeActionGroup;
  QAction *iModeAction[IpeOverlayFactory::Num];
  QComboBox *iMarkSize;
  DecimalSpinBox *iAbsMarkSize;
  QComboBox *iTextSize;
  DecimalSpinBox *iAbsTextSize;

  QActionGroup *iSnapActionGroup;
  QAction *iSnapAction[6];
  QComboBox *iGridSize;
  IpeSpinBox *iAbsGridSize;
  QComboBox *iAngleSize;
  DecimalSpinBox *iAbsAngleSize;

  LayerBox *iLayerList;
  QListWidget *iBookmarks;
  bool iBookmarksShown;

  QLabel *iMouse;
  IpeVector iMouseBase;
  int iMouseIn;

  QAction *iPasteBitmapAction;

  //! The most recently used file format on saving
  QString iSaveFilter;

  // Information about the document

  //! The document's filename (null if not yet saved).
  QString iFileName;
  //! The document.
  IpeDocument *iDoc;
  //! The currently displayed page no.
  int iPno;
  //! The current view.
  int iVno;
  //! The current layer.
  int iLayer;
  //! Whether latex needs to be run.
  bool iNeedLatex;
  //! Displaying attributes absolute (or symbolic)
  bool iAbsoluteAttributes;
  //! Displaying absolute snap sizes.
  bool iAbsoluteSnapping;
  //! The attributes set in the UI.
  IpeAllAttributes iAttributes;
  //! Snapping state
  IpeSnapData iSnapData;
  IpeFixed iSnapAngle;
  //! Current Ipe mode.
  int iMode;
  //! All symbolic attributes in style sheet
  IpeAttributeSeq iSyms[IpeAttribute::ETemplate + 1];

  //! Undo stack
  IpeUndoStack iUndo;
  // Ipelets
  //! Ipelet repository
  IpeletMaster iIpeletMaster;
  bool iIpeletNoUndo;

  // Key assignment
  static std::map<IpeString, IpeString> keyAssignment;
};

// --------------------------------------------------------------------
#endif
